@@ -4,6 +4,9 @@
 #include <vector>
 #include <unordered_map>
 #include <map>
+#include <memory>
+#include "mvt.hpp"
+#include "milo/dtoa_milo.h"
 
 enum attribute_op {
 	op_sum,
@@ -22,10 +25,13 @@ struct accum_state {
 };
 
 struct serial_val;
+struct key_pool;
 
 void set_attribute_accum(std::unordered_map<std::string, attribute_op> &attribute_accum, std::string name, std::string type);
 void set_attribute_accum(std::unordered_map<std::string, attribute_op> &attribute_accum, const char *arg, char **argv);
-void preserve_attribute(attribute_op const &op, const std::string &key, serial_val const &val, std::vector<std::string> &full_keys, std::vector<serial_val> &full_values, std::unordered_map<std::string, accum_state> &attribute_accum_state);
+
+void preserve_attribute(attribute_op const &op, std::string const &key, serial_val const &val, std::vector<std::shared_ptr<std::string>> &full_keys, std::vector<serial_val> &full_values, std::unordered_map<std::string, accum_state> &attribute_accum_state, key_pool &key_pool);
+void preserve_attribute(attribute_op const &op, std::string const &key, mvt_value const &val, std::vector<std::shared_ptr<std::string>> &full_keys, std::vector<mvt_value> &full_values, std::unordered_map<std::string, accum_state> &attribute_accum_state, key_pool &key_pool);
 
 extern std::map<std::string, attribute_op> numeric_operations;
 
